#
# ARMES 2D
#

Version = '1.1'
Date = 'Nov 2012'
Author = 'Luca Fini'

"""
This file contains some procedures which allows to
simulate some aspects of the Armes dimostrator

The file includes a simple test main program which can
be used to do some simple tests.

To run the test use:

   python armes2d.py

Prerequisites: numpy, scipy, matplotlib

The code has been tested in a Linux environment with:

Python Version 2.7
NumPy Version 1.6.1
SciPy Version 0.9.0
MatplotLib Version 1.1.1
"""

# The following four parameters provide values for heuristic optimization

XGAIN_OPT=1.0
XOFFSET_OPT=0.0

#XGAIN_OPT=1.418
#XOFFSET_OPT=-0.168

YGAIN_OPT=1.0
YOFFSET_OPT=0.0

import numpy as np
import os
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

import scipy.optimize

class ArmSizeError(Exception):
    pass

class ArmShiftError(Exception):
    pass

class ArmQuantError(Exception):
    pass

def gaussian(height, center_x, center_y, width_x, width_y):
    """Returns a gaussian function with the given parameters.
       Used by the fitting procedure"""
    width_x = float(width_x)
    width_y = float(width_y)
    return lambda x,y: height*np.exp(
                -(((center_x-x)/width_x)**2+((center_y-y)/width_y)**2)/2.)


def moments(data):
    """Returns (height, x, y, width_x, width_y)
    the gaussian parameters of a 2D distribution by calculating its moments.
    It is used by the fitting procedure to extimate initial values """
    total = data.sum()
    X, Y = np.indices(data.shape)
    x = (X*data).sum()/total
    y = (Y*data).sum()/total
    col = data[:, int(y)]
    width_x = np.sqrt(np.abs((np.arange(col.size)-y)**2*col).sum()/col.sum())
    row = data[int(x), :]
    width_y = np.sqrt(np.abs((np.arange(row.size)-x)**2*row).sum()/row.sum())
    height = data.max()
    return height, x, y, width_x, width_y

def fitgaussian(data):
    """Returns (height, x, y, width_x, width_y)
    the gaussian parameters of a 2D distribution found by a fit"""
    params = moments(data)
    errorfunction = lambda p: np.ravel(gaussian(*p)(*np.indices(data.shape)) - data)
    p, success = scipy.optimize.leastsq(errorfunction, params)
    return p

def gaussian2d(size,offset=0,width=10):
    '''Returns an array with gaussian data.
          size: size of square array, or (sx,sy)
          offset=center offset, or (ofx,ofy)
          width=circular width, or (wx,wy)'''
    if isinstance(size,(tuple,list)):
        sizex,sizey=int(size[0]),int(size[1])
    else:
        sizex=sizey=int(size)
    if isinstance(offset,(tuple,list)):
        ofstx,ofsty=float(offset[0]),float(offset[1])
    else:
        ofstx=ofsty=float(offset)
    if isinstance(width,(tuple,list)):
        widthx,widthy=float(width[0]),float(width[1])
    else:
        widthx=widthy=float(width)
    center=float(size)*0.5
    center_x=center+ofstx
    center_y=center+ofsty

    x=np.arange(sizex,dtype=np.float32)
    y=np.arange(sizey,dtype=np.float32).reshape((-1,1))

#  return np.exp(-4*np.log(2)*(((x-center_x)/widthx)**2 +((y-center_y)/widthy)**2))
    return np.exp(-0.5*(((x-center_x)/widthx)**2 +((y-center_y)/widthy)**2))

def tukey2d(size,alpha=0.5):
    """Returns a tukey 2D function (used for data apodization)"""
    size=int(size)

    fc=1.0/alpha
    step=2./float(size)
    s1=1.0-alpha


    def tk_kern(x):
        r=abs(x)
        if r > s1:
            return 0.5*(np.cos(np.pi*((r-1.)*fc+1))) +0.5
        else:
            return 1.0

    X,Y = np.mgrid[-1.:1.:step,-1:1:step]
    p=np.vectorize(tk_kern)
    h=p(X)*p(Y)
    return h

def quantization(img,nbits=10,saturation=0.9):
    '''Returns a quantized version of an image
       img: image to quantize. Modified on place.
       nbits: number of bits [8..16]
       saturation: saturation guard

  The saturation coefficient takes into account the real
  case where the input data range is kept somwhat greather than
  the maximum pixel value, to avoid saturation.'''

    if nbits<8 or nbits>16:
        print "Illegal quantization. Ignored."
        return
    coeff=2.0**int(nbits)*float(saturation)/img.data.max()
    img.data=np.floor(img.data*coeff)
    return

def addnoise(data,rms=0.2):
    "Adds noise as an rms percentage of signal"
    dc=np.sum(data*data)/len(data)
    d1=data-dc
    rmsdata=np.sqrt(sum(d1*d1)/len(data))
    rmsnoise=rmsdata*rms*8
    noise=np.random.uniform(0,rmsnoise,size=len(data))
    return data+noise

def crosscorr(l,r):
    """
  Computes the cross correlation using FFT/iFFT.
  """
    lf=np.fft.fft2(l)
    rf=np.fft.fft2(r)
    lf=np.conj(lf)
    iff=np.absolute(np.fft.ifft2(lf*rf))
    return np.fft.fftshift(iff)          # fftshift moves zero to center pixels

def findmax_gauss(data):
    '''
  Finds the maximum of a curve.

  Method: evaluates the fitting gaussian

  Returns: h, x, y, wx, wy
  '''

    return fitgaussian(data)

def computeAndShow(img,sx,sy,sub=20,quant=0,apodization=0.0,
                   XGAIN=1.0, XOFFSET=0.0,
                   YGAIN=1.0, YOFFSET=0.0,
                   showim=True,showcc=True,showres=True):
    '''This procedure performs the full simulation:

       1. Generates two shifted images (by interpolation)
       2. optionally applies quantization
       3. Optionally applies apodization
       4. Computes the cross correlation
       5. Fits a gaussian on the central part of the cross correlation matrix

       It uses matplotlib to optionally show results.
  '''

    i1,i2=twoimages(img,sx,sy)

    if quant:
        quantization(i1,quant)
        quantization(i2,quant)

    if apodization>0.0:
        apomat=tukey2d(i1.data.shape[0],alpha=apodization)
        i1.data=i1.data*apomat
        i2.data=i2.data*apomat

    if showim:
        plt.imshow(img,cmap='gray')
        l1=Line2D([i1.rect[0],i1.rect[1],i1.rect[1],i1.rect[0],i1.rect[0]],
                  [i1.rect[2],i1.rect[2],i1.rect[3],i1.rect[3],i1.rect[2]], color='red')
        ax=plt.gca()
        l2=Line2D([i2.rect[0],i2.rect[1],i2.rect[1],i2.rect[0],i2.rect[0]],
                  [i2.rect[2],i2.rect[2],i2.rect[3],i2.rect[3],i2.rect[2]],color='blue')
        ax.add_line(l1)
        ax.add_line(l2)
        plt.title('Original image and shifted views')
        plt.show()

    cc=crosscorr(i1.data,i2.data)            # calcola crosscorrelazione
    s0,d=cc.shape
    s0=float(s0)*0.5
    x0=int(s0-sub)
    x1=int(s0+sub)
    c0=cc[x0:x1,x0:x1]

    if showcc:
        plt.imshow(c0)
        plt.title('Cross correlation around zero')
        plt.show()
    h, x, y, wx, wy = findmax_gauss(c0)

    if showres:
        xi=int(x+0.5)
        yi=int(y+0.5)
        f=plt.figure()

        c=c0[xi,:]
        f.add_subplot(2,1,1)
        plt.plot(c,label='CC func.')
        g=gaussian(h, x, y, wx, wy)
        gy=g(x, np.arange(len(c)))
        plt.plot(gy,label='Fit')
        plt.title('X section')
        plt.legend()
        plt.grid()

        r=c0[:,yi]
        f.add_subplot(2,1,2)
        plt.plot(r,label='CC func.')
        plt.title('Y section')
        g=gaussian(h, x, y, wx, wy)
        gx=g(np.arange(len(r)),y)
        plt.plot(gx,label='Fit')
        plt.legend()
        plt.grid()

        plt.show()

                            # Adjust coordinates
    csx=float(sub)-y
    csy=float(sub)-x
                            # Remove systematic error by linear correction
    csx=csx*XGAIN+XOFFSET
    csy=csy*YGAIN+YOFFSET

    print "Applied: %f, %f   -   Computed: %f, %f" % (sx,sy,csx,csy)
    return (sx,sy),(csx,csy)    # Return values for further uses

class myImg(object):
    '''Defines an image object.
       myImg.data:  Actual image (numpy array)
       myImg.rect:  Absolute coordinates of the image corners '''
    def __init__(self,img):
        self.data=img
        self.rect=[]
    def set_rect(self,x0,x1,y0,y1):
        self.rect=[x0,x1,y0,y1]

def twoimages(img,sx,sy):
    '''Generates two 1024x1024 images extracting two shifted subimages from the
  central part of the input image'''

    from scipy.ndimage import map_coordinates
    x0_1=(img.shape[0]-1024 - sx)*0.5
    x1_1=x0_1+1023.5
    y0_1=(img.shape[1]-1024 - sy)*0.5
    y1_1=y0_1+1023.5
    xi=np.mgrid[y0_1:y1_1:1.0,x0_1:x1_1:1.0]  # swapped axes due to mesh generation details
    i1=myImg(map_coordinates(img,xi,order=1,mode='nearest'))
    i1.set_rect(x0_1,x1_1,y0_1,y1_1)

    x0_2=x0_1+sx
    x1_2=x1_1+sx
    y0_2=y0_1+sy
    y1_2=y1_1+sy
    xi=np.mgrid[y0_2:y1_2:1.0,x0_2:x1_2:1.0]  # swapped axes due to mesh generation details
    i2=myImg(map_coordinates(img,xi,order=1,mode='nearest'))
    i2.set_rect(x0_2,x1_2,y0_2,y1_2)

    return i1,i2

# The following procedures are used for interactive data input
# From the test main program

def input_shift():
    while True:
        l=raw_input('Shift value(s) [-10..10]: ')
        if not l: return None
        ll=l.split()
        if len(ll)<2: ll.append('0')
        try:
            sx,sy=float(ll[0]),float(ll[1])
        except:
            sx=sy=100001
        if sx<-10 or sx>10 or sy<-10 or sy>10:
            print 'Illegal shift (%s).' % l.strip()
        else:
            break
    return sx,sy

def input_quantization():
    nquant=0
    while True:
        l=raw_input('Quantization nbits [8-16, 0: no quantization]: ')
        if not l: break
        try:
            n=int(l)
        except:
            n=100
        if n==0: break
        if n<8 or n>16:
            print 'Illegal quantization (%d).' % n
        else:
            nquant=n
            break
    return nquant

def input_apo():
    apo=0.0
    while True:
        l=raw_input('Apodization [0..0.5]? ')
        if not l: break
        try:
            apo=float(l)
        except:
            apo=100.
        if apo<0.0 or apo>0.5:
            print "Illegal apodization value"
        else:
            break
    return apo

def input_neighbor():
    neighb=0
    while True:
        l=raw_input('Fitting neighborhood [5..30] [Default: 20]? ')
        if not l:
            neighb=20
            break
        try:
            neighb=int(l)
        except:
            neighb=100.
        if neighb<5 or neighb>30:
            print "Illegal neighborhood value"
        else:
            break
    return neighb

def input_image():
    while True:
        print
        im_file=raw_input('Image file name: ')
        if not im_file: return None
        im_file=os.path.abspath(im_file)
        try:
            img=scipy.misc.imread(im_file)
        except:
            print 'Cannot open file',im_file
            continue
        if len(img.shape) != 2:
            print 'Image must be greyscale, not color!'
            continue
        if img.shape[0]>=1100 and img.shape[1]>=1100:
            break
        else:
            print 'You need at least an image 1100x1100 pixels'
    return img



#
# Routines named test_nnn are specific tests called by
# the main program
#

def test_1():
    "Quick and dirty test with spikes"
    img=gaussian2d(1100,width=5)
    computeAndShow(img,3.33,-1.235)

def test_2():
    "Generates two images with a shifted peak and computes the shift"

    sx,sy=input_shift()
    if sx is None: return

    nquant=input_quantization()

    img=gaussian2d(1100,width=5)

    computeAndShow(img,sx,sy,quant=nquant)

def test_3():
    'Gets two shifted subimages from an image and computes the shift'

    print 'An empty answer to most prompts terminates the test'

    img=input_image()
    if img is None: return

    nquant=input_quantization()
    sx,sy=input_shift()
    if sx is None: return

#  apo=input_apo()
    apo=0

    neighb=input_neighbor()

    computeAndShow(img,sx,sy,sub=neighb,quant=nquant,apodization=apo)

def test_4():
    'Test different shift values on given image'

    from scipy import stats

    print 'An empty answer to most prompts terminates the test'

    img=input_image()
    if img is None: return

    apo=input_apo()

    ssx=[]      # Applied shifts
    csx=[]      # computed X shifts
    points=map(lambda x: float(x)*0.1,range(-14,15))
    for sx in points:
        r=computeAndShow(img,sx,0.0,sub=10, apodization=apo,
                         XGAIN=XGAIN_OPT, XOFFSET=XOFFSET_OPT,
                         showim=False,showcc=False,showres=False)
        ssx.append(r[0][0])
        csx.append(r[1][0])

    ssx=np.array(ssx)      # Applied shifts
    csx=np.array(csx)      # computed X shifts

    plt.plot(points,ssx,label='Applied')
    plt.plot(points,csx,'+',label='Computed')
    plt.legend()
    plt.grid()
    plt.show()

# err=csx-ssx
# plt.plot(points,err)
# plt.title('Error')
# plt.grid()
# plt.show()

# Heuristic parameters evaluation
    m, c = stats.linregress(points,csx)[0:2]

# Compute heuristics corrections
    xgain=1./m
    xoffset= -c/m

    print
    print "   Current Heuristics:  XGAIN=%f   XOFFSET=%f" % (XGAIN_OPT,XOFFSET_OPT)
    print " Suggested Heuristics:  XGAIN=%f   XOFFSET=%f" % (xgain,xoffset)

#
# Main program. Launches the subtests upon user's selection
#
def main():
    gb=globals()
    tests=filter(lambda x: (x+'      ')[:5]=='test_', gb)
    tests.sort()
    while True:
        print
        print 'Available tests:'
        for k in tests:
            print '  %s'%k[5:],'-',gb[k].__doc__
        print
        n=raw_input('Test Number [0 to exit]: ')
        try:
            n=int(n)
        except:
            break
        if n<=0: break
        eval('test_%d()'%n)

if __name__=='__main__': main()
